package com.chelvc.framework.redis.queue;

import java.util.AbstractQueue;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.function.Function;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.common.util.AssertUtils;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.common.util.StringUtils;
import com.chelvc.framework.redis.context.RedisContextHolder;
import com.google.common.collect.Maps;
import lombok.NonNull;
import org.springframework.data.redis.core.RedisTemplate;

/**
 * Redis队列实现
 *
 * @param <E> 元素类型
 * @author Woody
 * @date 2024/8/19
 */
public class RedisQueue<E> extends AbstractQueue<E> {
    /**
     * 队列名称/实例映射表
     */
    @SuppressWarnings("rawtypes")
    private static final Map<String, RedisQueue> INSTANCES = Maps.newConcurrentMap();

    private final String name;

    public RedisQueue(@NonNull String name) {
        String profile = ApplicationContextHolder.getProfile();
        this.name = StringUtils.isEmpty(profile) ? name : (profile + "-" + name);
    }

    /**
     * 集合迭代器实现
     */
    private class Iter implements Iterator<E> {
        private E element;
        private int index = 0;
        private final int size;

        public Iter(int size) {
            AssertUtils.check(size > -1, () -> "size must be greater than -1");
            this.size = size;
        }

        @Override
        public boolean hasNext() {
            return this.index < this.size;
        }

        @Override
        public E next() {
            return this.element = template().opsForList().index(name, this.index++);
        }

        @Override
        public void remove() {
            if (this.element != null) {
                template().opsForList().remove(name, 1, this.element);
                this.element = null;
            }
        }
    }

    /**
     * 构建消息队列
     *
     * @param name 队列名称
     * @param <E>  消息类型
     * @return 队列实例
     */
    public static <E> RedisQueue<E> build(@NonNull String name) {
        return build(name, RedisQueue::new);
    }

    /**
     * 构建消息队列
     *
     * @param name     队列名称
     * @param function 队列构建函数
     * @param <E>      消息类型
     * @return 队列实例
     */
    @SuppressWarnings("unchecked")
    public static <E> RedisQueue<E> build(@NonNull String name, @NonNull Function<String, RedisQueue<E>> function) {
        RedisQueue<E> queue = INSTANCES.get(name);
        return queue == null ? INSTANCES.computeIfAbsent(name, function) : queue;
    }

    /**
     * 获取RedisTemplate实例
     *
     * @return RedisTemplate实例
     */
    protected RedisTemplate<String, E> template() {
        return RedisContextHolder.getDefaultTemplate();
    }

    @Override
    public Iterator<E> iterator() {
        return new Iter(this.size());
    }

    @Override
    public int size() {
        Long size = this.template().opsForList().size(this.name);
        return size == null ? 0 : size.intValue();
    }

    @Override
    public boolean offer(E e) {
        if (e == null) {
            return false;
        }

        Long count = this.template().opsForList().rightPush(this.name, e);
        return count != null && count > 0;
    }

    @Override
    public E poll() {
        return this.template().opsForList().leftPop(this.name);
    }

    @Override
    public E peek() {
        return this.template().opsForList().index(this.name, 0);
    }

    @Override
    public void clear() {
        this.template().delete(this.name);
    }

    @Override
    @SuppressWarnings("unchecked")
    public boolean addAll(Collection<? extends E> c) {
        if (ObjectUtils.isEmpty(c)) {
            return false;
        }

        Long count = this.template().opsForList().rightPushAll(this.name, (Collection<E>) c);
        return count != null && count > 0;
    }

    @Override
    public Object[] toArray() {
        List<E> values = this.template().opsForList().range(this.name, 0, -1);
        return ObjectUtils.ifNull(values, Collections::emptyList).toArray();
    }

    @Override
    public <T> T[] toArray(T[] a) {
        List<E> values = this.template().opsForList().range(this.name, 0, -1);
        return ObjectUtils.ifNull(values, Collections::emptyList).toArray(a);
    }

    @Override
    public String toString() {
        List<E> values = this.template().opsForList().range(this.name, 0, -1);
        return ObjectUtils.ifNull(values, Collections::emptyList).toString();
    }
}
